==> runtime-simulator-0.8.0.tgz: package/src/controller.ts <==
/**
 * @license
 * Copyright 2022 Dynatrace LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { yellow } from 'chalk';
import type { ChildProcessWithoutNullStreams } from 'child_process';
import { spawn } from 'child_process';
import { EventEmitter } from 'events';
import { Socket } from 'net';
import { join } from 'path';
import { serialize, deserialize } from 'v8';
import type { RpcResponse } from './rpc';
import type { Task } from './task';

/** Serverless worker server state */
const enum State {
  none,
  initializing,
  ready,
}

const readyListener = new EventEmitter();

let childProcess: ChildProcessWithoutNullStreams | undefined;
let port = 0;
let state = State.none;

/** Initialize the serverless worker server */
export async function initialize(): Promise<void> {
  state = State.initializing;

  // We must keep the `NODE_OPTIONS` from the environment here so that debugging with VS Code works
  // VS Code passes the following command line flags via `NODE_OPTIONS`:
  // --require "/Applications/Visual Studio Code.app/Contents/Resources/app/extensions/ms-vscode.js-debug/src/bootloader.bundle.js" --enable-source-maps --inspect-publish-uid=http
  const opts =
    (process.env.NODE_OPTIONS ?? '') +
    ' --experimental-vm-modules --no-warnings --enable-source-maps';

  // Spawn a new child process with the serverless worker server
  childProcess = spawn(process.execPath, [join(__dirname, 'server.js')], {
    env: {
      ...process.env,
      NODE_OPTIONS: opts,
      FORCE_COLOR: '1',
    },
  });
  childProcess!.stderr.on('data', (data: Buffer) => {
    console.error(yellow('[runtime-simulator STDERR]'), data.toString());
  });

  return new Promise<void>((resolve, reject) => {
    const dataListener = (data: Buffer) => {
      // Wait for the serverless worker server to become ready
      if (data.toString().startsWith('Listening on port')) {
        port = Number.parseInt(
          data.toString().substring('Listening on port'.length).trim(),
        );

        clearTimeout(timeoutHandle);
        childProcess!.stdout.removeListener('data', dataListener);

        resolve();
      }
    };

    const timeoutHandle = setTimeout(() => {
      childProcess!.stdout.removeListener('data', dataListener);

      reject(
        new Error('Failed to start the serverless worker server: Timeout'),
      );
    }, 8000);

    childProcess!.stdout.on('data', dataListener);
  }).then(() => {
    state = State.ready;

    childProcess!.stdout.on('data', (data: Buffer) => {
      // Pipe the stdout output to the Console API and remove the last newline character
      console.log(data.toString().replace(/\n$/, ''));
    });

    if (readyListener.listenerCount('') > 0) {
      readyListener.emit('');
    }
  });
}

/**
 * Dispatch a serverless worker task.
 *
 * Creates a new TCP socket connection for every serverless execution to the
 * serverless worker server. Once the response from the server has been read,
 * the worker server will close the socket with FIN.
 */
export async function dispatch<T>(task: Task): Promise<T> {
  if (state === State.none) {
    throw new Error(
      'Failed to dispatch serverless worker task: controller not initialized',
    );
  } else if (state === State.initializing) {
    await new Promise<void>((resolve) => {
      readyListener.once('', () => resolve());
    });
  }

  // Create a new socket
  const socket = new Socket();

  return new Promise<void>((resolve, reject) => {
    const timeoutHandle = setTimeout(() => {
      reject(
        new Error('Failed to connect to the serverless worker server: Timeout'),
      );
    }, 8000);

    // Connect to the remote server
    socket.connect(port, 'localhost', () => {
      clearTimeout(timeoutHandle);

      resolve();
    });
  })
    .then(() => {
      return new Promise<T>((resolve, reject) => {
        const chunks: Buffer[] = [];

        /** Data event listener */
        function dataListener(chunk: Buffer) {
          chunks.push(chunk);
        }

        socket.on('data', dataListener);
        socket.once('close', () => {
          socket.removeListener('data', dataListener);

          // Deserialize the RPC response
          const rpcResponse: RpcResponse<T> = deserialize(
            Buffer.concat(chunks),
          );
          if (Object.prototype.hasOwnProperty.call(rpcResponse, 'result')) {
            resolve(rpcResponse.result!);
          } else {
            reject(deserialize(rpcResponse.error!));
          }
        });

        try {
          const data = serialize(task);

          socket.write(
            Buffer.concat([
              // Add the content length
              Buffer.from(data.byteLength.toString()),
              data,
            ]),
            (error: Error | undefined) => {
              if (error) {
                reject(error);
              }
            },
          );
        } catch (error) {
          reject(error);
        }
      });
    })
    .finally(() => {
      if (socket.readable) {
        socket.end();
      }
    });
}

process.once('exit', () => {
  if (childProcess) {
    childProcess.kill();
  }
});

==> runtime-simulator-0.9.1.tgz: package/src/controller.ts <==
/**
 * @license
 * Copyright 2022 Dynatrace LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import { yellow } from 'chalk';
import type { ChildProcessWithoutNullStreams } from 'child_process';
import { spawn } from 'child_process';
import { EventEmitter } from 'events';
import { Socket } from 'net';
import { join } from 'path';
import { serialize, deserialize } from 'v8';
import type { RpcResponse } from './rpc';
import type { Task } from './task';

/** Serverless worker server state */
const enum State {
  none,
  initializing,
  ready,
}

const readyListener = new EventEmitter();

let childProcess: ChildProcessWithoutNullStreams | undefined;
let port = 0;
let state = State.none;

/** Initialize the serverless worker server */
export async function initialize(
  serverWorkerPath?: string,
): Promise<ChildProcessWithoutNullStreams | undefined> {
  state = State.initializing;

  // We must keep the `NODE_OPTIONS` from the environment here so that debugging with VS Code works
  // VS Code passes the following command line flags via `NODE_OPTIONS`:
  // --require "/Applications/Visual Studio Code.app/Contents/Resources/app/extensions/ms-vscode.js-debug/src/bootloader.bundle.js" --enable-source-maps --inspect-publish-uid=http
  const opts =
    (process.env.NODE_OPTIONS ?? '') +
    ' --experimental-vm-modules --no-warnings --enable-source-maps';

  // Spawn a new child process with the serverless worker server
  childProcess = spawn(
    process.execPath,
    [serverWorkerPath || join(__dirname, 'server.js')],
    {
      env: {
        ...process.env,
        NODE_OPTIONS: opts,
        FORCE_COLOR: '1',
      },
    },
  );
  childProcess!.stderr.on('data', (data: Buffer) => {
    console.error(yellow('[runtime-simulator STDERR]'), data.toString());
  });

  return new Promise<void>((resolve, reject) => {
    const dataListener = (data: Buffer) => {
      // Wait for the serverless worker server to become ready
      if (data.toString().startsWith('Listening on port')) {
        port = Number.parseInt(
          data.toString().substring('Listening on port'.length).trim(),
        );

        clearTimeout(timeoutHandle);
        childProcess!.stdout.removeListener('data', dataListener);

        resolve();
      }
    };

    const timeoutHandle = setTimeout(() => {
      childProcess!.stdout.removeListener('data', dataListener);

      reject(
        new Error('Failed to start the serverless worker server: Timeout'),
      );
    }, 8000);

    childProcess!.stdout.on('data', dataListener);
  }).then(() => {
    state = State.ready;

    childProcess!.stdout.on('data', (data: Buffer) => {
      // Pipe the stdout output to the Console API and remove the last newline character
      console.log(data.toString().replace(/\n$/, ''));
    });

    if (readyListener.listenerCount('') > 0) {
      readyListener.emit('');
    }

    return childProcess;
  });
}

/**
 * Dispatch a serverless worker task.
 *
 * Creates a new TCP socket connection for every serverless execution to the
 * serverless worker server. Once the response from the server has been read,
 * the worker server will close the socket with FIN.
 */
export async function dispatch<T>(task: Task): Promise<T> {
  if (state === State.none) {
    throw new Error(
      'Failed to dispatch serverless worker task: controller not initialized',
    );
  } else if (state === State.initializing) {
    await new Promise<void>((resolve) => {
      readyListener.once('', () => resolve());
    });
  }

  // Create a new socket
  const socket = new Socket();

  return new Promise<void>((resolve, reject) => {
    const timeoutHandle = setTimeout(() => {
      reject(
        new Error('Failed to connect to the serverless worker server: Timeout'),
      );
    }, 8000);

    // Connect to the remote server
    socket.connect(port, 'localhost', () => {
      clearTimeout(timeoutHandle);

      resolve();
    });
  })
    .then(() => {
      return new Promise<T>((resolve, reject) => {
        const chunks: Buffer[] = [];

        /** Data event listener */
        function dataListener(chunk: Buffer) {
          chunks.push(chunk);
        }

        socket.on('data', dataListener);
        socket.once('close', () => {
          socket.removeListener('data', dataListener);

          // Deserialize the RPC response
          const rpcResponse: RpcResponse<T> = deserialize(
            Buffer.concat(chunks),
          );
          if (Object.prototype.hasOwnProperty.call(rpcResponse, 'result')) {
            resolve(rpcResponse.result!);
          } else {
            reject(deserialize(rpcResponse.error!));
          }
        });

        try {
          const data = serialize(task);

          socket.write(
            Buffer.concat([
              // Add the content length
              Buffer.from(data.byteLength.toString()),
              data,
            ]),
            (error: Error | undefined) => {
              if (error) {
                reject(error);
              }
            },
          );
        } catch (error) {
          reject(error);
        }
      });
    })
    .finally(() => {
      if (socket.readable) {
        socket.end();
      }
    });
}

process.once('exit', () => {
  if (childProcess) {
    childProcess.kill();
  }
});
